from . import piece
Piece = piece.Piece
from . import puzzle
Puzzle = puzzle.Puzzle
QApp = puzzle.QApp

from . import param
from . import readout
from . import action
from . import parse
from . import threads